@@ -2,21 +2,22 @@
  * Copyright (c) 2000 Apple Computer, Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
- *
- * The contents of this file constitute Original Code as defined in and
- * are subject to the Apple Public Source License Version 1.1 (the
- * "License").  You may not use this file except in compliance with the
- * License.  Please obtain a copy of the License at
- * http://www.apple.com/publicsource and read it before using this file.
- *
- * This Original Code and all software distributed under the License are
- * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
+ * 
+ * This file contains Original Code and/or Modifications of Original Code
+ * as defined in and that are subject to the Apple Public Source License
+ * Version 2.0 (the 'License'). You may not use this file except in
+ * compliance with the License. Please obtain a copy of the License at
+ * http://www.opensource.apple.com/apsl/ and read it before using this
+ * file.
+ * 
+ * The Original Code and all software distributed under the License are
+ * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
  * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
  * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
- * License for the specific language governing rights and limitations
- * under the License.
- *
+ * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
+ * Please see the License for the specific language governing rights and
+ * limitations under the License.
+ * 
  * @APPLE_LICENSE_HEADER_END@
  */
 
@@ -69,9 +70,9 @@
 #include "../../../Family/if_ppp.h"
 #include "../../../Family/ppp_domain.h"
 #include "../PPPoE-extension/PPPoE.h"
-#include "../../../ppp/pppd/pppd.h"
-#include "../../../ppp/pppd/fsm.h"
-#include "../../../ppp/pppd/lcp.h"
+#include "../../../Helpers/pppd/pppd.h"
+#include "../../../Helpers/pppd/fsm.h"
+#include "../../../Helpers/pppd/lcp.h"
 
 
 /* -----------------------------------------------------------------------------
@@ -84,18 +85,26 @@
 
 #define PPPOE_NKE	"PPPoE.kext"
 
+// PPPoE error codes (bits 8..15 of last cause key)
+#define EXIT_PPPoE_NOSERVER  		1
+#define EXIT_PPPoE_NOSERVICE  		2
+#define EXIT_PPPoE_NOAC 		3
+#define EXIT_PPPoE_NOACSERVICE 		4
+#define EXIT_PPPoE_CONNREFUSED 		5
+
 /* -----------------------------------------------------------------------------
  Forward declarations
 ----------------------------------------------------------------------------- */
-void pppoe_device_check();
+void pppoe_process_extra_options();
 void pppoe_check_options();
-int pppoe_connect();
+int pppoe_connect(int *errorcode);
 void pppoe_disconnect();
-void pppoe_close_fds();
+void pppoe_close();
 void pppoe_cleanup();
 int pppoe_establish_ppp(int);
 void pppoe_wait_input();
 void pppoe_disestablish_ppp(int);
+void pppoe_link_down(void *arg, int p);
 
 static int pppoe_dial();
 static int pppoe_listen();
@@ -116,9 +125,9 @@ static bool 	loopback = 0;			/* loop back mode */
 static bool 	noload = 0;			/* don't load the kernel extension */
 static char	*service = NULL; 		/* service selection to use */
 static char	*access_concentrator = NULL; 	/* access concentrator to connect to */
-static char	*interface = "en0"; 		/* ethernet interface to use */
 static int	retrytimer = 0; 		/* retry timer (default is 3 seconds) */
 static int	connecttimer = 65; 		/* bump the connection timer from 20 to 65 seconds */
+static bool	linkdown = 0; 			/* flag set when we receive link down event */
 
 extern int kill_link;
 
@@ -134,8 +143,6 @@ option_t pppoe_options[] = {
       "Don't try to load the PPPoE kernel extension", 1 },
     { "pppoemode", o_string, &mode,
       "Configure configuration mode [connect, listen, answer]" },
-    { "pppoeinterface", o_string, &interface,
-      "Ethernet interface to use" },
     { "pppoeconnecttimer", o_int, &connecttimer,
       "Connect timer for outgoing call (default 65 seconds)" },
     { "pppoeretrytimer", o_int, &retrytimer,
@@ -152,40 +159,54 @@ int start(CFBundleRef ref)
  
     bundle = ref;
     CFRetain(bundle);
-    
-    // add the socket specific options
-    add_options(pppoe_options);
-    
+        
     // hookup our socket handlers
-    dev_device_check_hook = pppoe_device_check;
-    dev_wait_input_hook = pppoe_wait_input;
-    dev_check_options_hook = pppoe_check_options;
-    dev_connect_hook = pppoe_connect;
-    dev_disconnect_hook = pppoe_disconnect;
-    dev_cleanup_hook = pppoe_cleanup;
-    dev_close_fds_hook = pppoe_close_fds;
-    dev_establish_ppp_hook = pppoe_establish_ppp;
-    dev_disestablish_ppp_hook = pppoe_disestablish_ppp;
+    bzero(the_channel, sizeof(struct channel));
+    the_channel->options = pppoe_options;
+    the_channel->process_extra_options = pppoe_process_extra_options;
+    the_channel->wait_input = pppoe_wait_input;
+    the_channel->check_options = pppoe_check_options;
+    the_channel->connect = pppoe_connect;
+    the_channel->disconnect = pppoe_disconnect;
+    the_channel->cleanup = pppoe_cleanup;
+    the_channel->close = pppoe_close;
+    the_channel->establish_ppp = pppoe_establish_ppp;
+    the_channel->disestablish_ppp = pppoe_disestablish_ppp;
+    // use the default config functions
+    the_channel->send_config = generic_send_config;
+    the_channel->recv_config = generic_recv_config;
+    
+    add_notifier(&link_down_notifier, pppoe_link_down, 0);
     return 0;
 }
 
+/* -----------------------------------------------------------------------------
+----------------------------------------------------------------------------- */
+void pppoe_link_down(void *arg, int p)
+{
+    linkdown = 1;
+}
+
 /* ----------------------------------------------------------------------------- 
 work out which device we are using and read its options file
 ----------------------------------------------------------------------------- */
-void pppoe_device_check()
+void pppoe_process_extra_options()
 {
     int 		error = EXIT_OPTION_ERROR;
     int 		len, s;
     struct ifreq 	ifr;
 
+    if (!device)
+        device = "en0";
+        
     s = socket(PF_SYSTEM, SOCK_RAW, SYSPROTO_EVENT);
     if (s >= 0) {
         
-        len = strlen(interface);
+        len = strlen(device);
         if (len <= sizeof(ifr.ifr_name)) {
 
             bzero(&ifr, sizeof(ifr));
-            bcopy(interface, ifr.ifr_name, len);
+            bcopy(device, ifr.ifr_name, len);
             if (ioctl(s, SIOCGIFFLAGS, (caddr_t) &ifr) >= 0) {
                 // ensure that the device is UP
                 ifr.ifr_flags |= IFF_UP;
@@ -197,6 +218,11 @@ void pppoe_device_check()
         close(s);
     }
         
+    if (!strcmp(mode, MODE_ANSWER)) {
+        // make sure we get a file descriptor > 2 so that pppd can detach and close 0,1,2
+        sockfd = dup(0);
+    }
+
     if (error)
         exit(error);
 }
@@ -218,8 +244,9 @@ void pppoe_wait_input()
 {
    
     if (sockfd != -1 && is_ready_fd(sockfd)) {
-        // it's OK to get a hangup during terminate phase
-        if (phase != PHASE_TERMINATE) {
+        // looks like we have been disconnected...
+        // the status is updated only if link is not already down
+        if (linkdown == 0) {
             notice("PPPoE hangup");
             status = EXIT_HANGUP;
         }
@@ -234,7 +261,7 @@ void pppoe_wait_input()
 get the socket ready to start doing PPP.
 That is, open the socket and run the connector
 ----------------------------------------------------------------------------- */
-int pppoe_connect()
+int pppoe_connect(int *errorcode)
 {
     char 	dev[32], name[MAXPATHLEN]; 
     int 	err = 0;  
@@ -245,32 +272,35 @@ int pppoe_connect()
 
     hungup = 0;
     kill_link = 0;
+    linkdown = 0;
 
-    /* open the socket */
-    sockfd = socket(PF_PPP, SOCK_DGRAM, PPPPROTO_PPPOE);
-    if (sockfd < 0) {
-        if (!noload) {
-            if (url = CFBundleCopyBundleURL(bundle)) {
-                name[0] = 0;
-                CFURLGetFileSystemRepresentation(url, 0, name, MAXPATHLEN - 1);
-                CFRelease(url);
-                strcat(name, "/");
-                if (url = CFBundleCopyBuiltInPlugInsURL(bundle)) {
-                    CFURLGetFileSystemRepresentation(url, 0, name + strlen(name), 
-                        MAXPATHLEN - strlen(name) - strlen(PPPOE_NKE) - 1);
+    if (strcmp(mode, MODE_ANSWER)) {
+        /* open the socket */
+        sockfd = socket(PF_PPP, SOCK_DGRAM, PPPPROTO_PPPOE);
+        if (sockfd < 0) {
+            if (!noload) {
+                if (url = CFBundleCopyBundleURL(bundle)) {
+                    name[0] = 0;
+                    CFURLGetFileSystemRepresentation(url, 0, name, MAXPATHLEN - 1);
                     CFRelease(url);
                     strcat(name, "/");
-                    strcat(name, PPPOE_NKE);
-                    if (!load_kext(name))
-                        sockfd = socket(PF_PPP, SOCK_DGRAM, PPPPROTO_PPPOE);
-                }	
+                    if (url = CFBundleCopyBuiltInPlugInsURL(bundle)) {
+                        CFURLGetFileSystemRepresentation(url, 0, name + strlen(name), 
+                            MAXPATHLEN - strlen(name) - strlen(PPPOE_NKE) - 1);
+                        CFRelease(url);
+                        strcat(name, "/");
+                        strcat(name, PPPOE_NKE);
+                        if (!load_kext(name))
+                            sockfd = socket(PF_PPP, SOCK_DGRAM, PPPPROTO_PPPOE);
+                    }	
+                }
+            }
+            if (sockfd < 0) {
+                error("Failed to open PPPoE socket: %m");
+                status = EXIT_OPEN_FAILED;
+                return -1;
             }
         }
-        if (sockfd < 0) {
-            error("Failed to open PPPoE socket: %m");
-            status = EXIT_OPEN_FAILED;
-            return -1;
-       }
     }
 
     if (loopback || debug) {
@@ -301,7 +331,7 @@ int pppoe_connect()
         }
     }
 
-    if (setsockopt(sockfd, PPPPROTO_PPPOE, PPPOE_OPT_INTERFACE, interface, strlen(interface))) {
+    if (setsockopt(sockfd, PPPPROTO_PPPOE, PPPOE_OPT_INTERFACE, device, strlen(device))) {
         error("PPPoE can't specify interface...\n");
         return errno;
     }
@@ -319,7 +349,11 @@ int pppoe_connect()
         fatal("PPPoE incorrect mode : '%s'", mode ? mode : "");
 
     if (err) {
-        status = EXIT_CONNECT_FAILED;
+        if (err != -2) {
+            if (err != -1)
+                devstatus = err;
+            status = EXIT_CONNECT_FAILED;
+        }
         return -1;
     }
     
@@ -344,7 +378,7 @@ void pppoe_disconnect()
 /* ----------------------------------------------------------------------------- 
 close the socket descriptors
 ----------------------------------------------------------------------------- */
-void pppoe_close_fds()
+void pppoe_close()
 {
 	if (sockfd >= 0) {
 		close(sockfd);
@@ -357,7 +391,7 @@ clean up before quitting
 ----------------------------------------------------------------------------- */
 void pppoe_cleanup()
 {
-    pppoe_close_fds();
+    pppoe_close();
 }
 
 /* ----------------------------------------------------------------------------- 
@@ -365,19 +399,25 @@ establish the socket as a ppp link
 ----------------------------------------------------------------------------- */
 int pppoe_establish_ppp(int fd)
 {
-    int 	x;
-
+    int x, new_fd;
+    
     if (ioctl(fd, PPPIOCATTACH, &x) < 0) {
         error("Couldn't attach socket to the link layer: %m");
         return -1;
     }
 
+    new_fd = generic_establish_ppp(fd);
+    if (new_fd == -1)
+        return -1;
+
+    /* add our pppoe socket to the select */
     add_fd(fd);
-    return 0;
+    
+    return new_fd;
 }
 
 /* ----------------------------------------------------------------------------- 
-establish the socket as a ppp link
+disestablish the socket as a ppp link
 ----------------------------------------------------------------------------- */
 void pppoe_disestablish_ppp(int fd)
 {
@@ -387,6 +427,8 @@ void pppoe_disestablish_ppp(int fd)
 
     if (ioctl(fd, PPPIOCDETACH, &x) < 0)
         error("Couldn't detach socket from link layer: %m");
+
+    generic_disestablish_ppp(fd);
 }
 
 /* -----------------------------------------------------------------------------
@@ -427,8 +469,31 @@ int pppoe_dial()
     if (service)
         strncpy(addr.pppoe_service, service, sizeof(addr.pppoe_service));
     if (connect(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_pppoe)) < 0) {
+        if (errno == EINTR)
+            return -2; // user cancelled
         error("PPPoE connection failed, %m");
-        return errno;
+        switch (errno) {
+            case EHOSTUNREACH:
+                if ((service && service[0]) && (access_concentrator && access_concentrator[0]))
+                    return EXIT_PPPoE_NOACSERVICE;
+                else if (service && service[0])
+                    return EXIT_PPPoE_NOSERVICE;
+                else if (access_concentrator && access_concentrator[0])
+                    return EXIT_PPPoE_NOAC;
+                return EXIT_PPPoE_NOSERVER;
+                
+            case ECONNREFUSED:
+                return EXIT_PPPoE_CONNREFUSED;
+
+            case ENXIO:
+                // Ethernet interface is detached
+                // fake a cancel to get a consistent
+                // error message with the HANGUP case
+                status = EXIT_HANGUP;
+                return -2; 
+
+        }
+        return -1;
     }
 
     info("PPPoE connection established.");
@@ -457,19 +522,19 @@ int pppoe_listen()
         strncpy(addr.pppoe_service, service, sizeof(addr.pppoe_service));
     if (bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_pppoe)) < 0) {
         error("PPPoE bind failed, %m");
-        return errno;
+        return -1;
     }
 
     if (listen(sockfd, 10) < 0) {
         printf("PPPoE listen failed, %m");
-        return errno;
+        return -1;
     }
 
     len = sizeof(addr);
     fd = accept(sockfd, (struct sockaddr *) &addr, &len);
     if (fd < 0) {
         error("PPPoE accept failed, %m");
-        return errno;
+        return -1;
     }
     
     close(sockfd);	// close the socket used for listening
